@@ -2,7 +2,7 @@ import traceback
 from getpass import getpass
 
 from seclea_ai.exceptions import AuthenticationError
-from seclea_ai.lib.seclea_utils.core import Transmission
+from seclea_ai.lib.seclea_utils.core.transmission import Transmission
 from .storage import Storage
 
 try:
