@@ -8,6 +8,9 @@ from ..object_manager import DatasetManager
 
 
 class TensorflowDatasetManager(DatasetManager):
+
+    framework = "tensorflow"
+
     def save_object(self, obj: tf.data.Dataset):
         tf.data.experimental.save(
             obj, self.object_file_path, compression=self.metadata.get("compression", "NONE")
@@ -21,3 +24,6 @@ class TensorflowDatasetManager(DatasetManager):
 
     def hash_object(self, obj) -> int:
         return hash(obj)
+
+    def get_outputs_info(self):
+        raise NotImplementedError("Dataset info detection is currently not supported for Tensorflow datasets.")
