@@ -3,6 +3,7 @@ from typing import Any, Dict
 
 import tensorflow as tf
 
+from ..enums import ApplicationType
 from ..object_manager import ModelManager
 
 
@@ -32,3 +33,9 @@ class TensorflowModelManager(ModelManager):
         :param model: The model
         """
         return json.loads(model.to_json())
+
+    @staticmethod
+    def get_application_type(model) -> ApplicationType:
+        # we just need to know the output dimensionality of the model for binary|multiclass
+        # TODO look up how regression works
+        return ApplicationType.UNKNOWN
