@@ -17,7 +17,7 @@ class TestFileCompressor(unittest.TestCase):
     def setUp(self) -> None:
         base_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
         self.folder_path = os.path.join(base_dir, "tests")
-        self.dataset = f"{self.folder_path}/example_files/filled_na_train.csv"
+        self.dataset = f"{self.folder_path}/data/filled_na_train.csv"
         self.df = pd.read_csv(self.dataset)
         self.series = self.df["DeviceType"]
 
